@@ -21,7 +21,7 @@ set -eu
 # Boston, MA 02110-1301 USA.                                             #
 ##########################################################################
 
-# Depends: debootstrap, rsync, python-minimal|python, procps, squashfs-tools, ltsp-server [i386]
+# Depends: debootstrap, rsync, python-minimal|python, procps, squashfs-tools, ltsp-server [i386], e2fsprogs
 
 cleanup() {
     for mnt in ${ROOT}dev/pts ${ROOT}dev/shm ${ROOT}.dev ${ROOT}dev \
@@ -30,7 +30,7 @@ cleanup() {
 	umount $mnt || true
     done
 
-    [ -n "$DEV" ] && losetup -d $DEV || true
+    [ -n "$DEV" ] && umount $DEV && losetup -d $DEV || true
     grep ${ROOT} /proc/mounts && return 1 || return 0
 }
 
@@ -56,11 +56,63 @@ subst_package() {
 }
 
 
+livefs_squash()
+{
+  squashsort="http://people.ubuntu.com/~tfheen/livesort/${FSS}.list.${TARGETARCH}"
+  #if wget -O livecd.${FSS}.sort ${squashsort} > /dev/null 2>&1; then
+  if false; then
+    echo "Using the squashfs sort list from ${squashsort}."
+  else
+    echo "Unable to fetch squashfs sort list; using a blank list."
+    : > livecd.${FSS}.sort
+  fi
+
+  # make sure there is no old squashfs idling around
+  rm -f livecd.${FSS}.squashfs
+
+  mksquashfs ${ROOT} livecd.${FSS}.squashfs -sort livecd.${FSS}.sort
+  chmod 644 livecd.${FSS}.squashfs
+}
+
+livefs_ext2()
+{
+  # Add 500MiB extra free space for first boot + ext3 journal
+  size=$(($(du -ks ${ROOT} | cut -f1) + (512000)))
+  MOUNTPOINT=$(mktemp -d)
+  DEV=$(losetup -f)
+  echo "Building ext2 filesystem."
+
+  # remove any stale filesystem images
+  rm -f livecd.${FSS}.ext?
+
+  # create an empty ext2 image and loop mount it
+  dd if=/dev/zero of=livecd.${FSS}.ext2 bs=1024 count=0 seek=$size
+  mke2fs -i 8192 -F livecd.${FSS}.ext2
+  mount -o loop=${DEV} livecd.${FSS}.ext2 ${MOUNTPOINT}
+
+  # copy chroot content to image
+  cp -a ${ROOT}/* ${MOUNTPOINT}
+
+  # clean up
+  umount ${MOUNTPOINT}
+  rm -rf ${MOUNTPOINT}
+  losetup -d $DEV || true
+  chmod 644 livecd.${FSS}.ext2
+}
+
+checkpoint()
+{
+  echo "===== $1 ====="
+  date
+}
+
 if [ $(id -u) != 0 ];then
   echo "must be run as root"
   exit 2
 fi
 
+checkpoint "Starting build"
+
 umask 022
 export TTY=unknown
 export TERM=vt100
@@ -72,6 +124,7 @@ PPAMIRROR=ppa.launchpad.net
 ARCH=$(dpkg --print-architecture)
 OPTMIRROR=
 INITRD_COMPRESSOR=lzma
+TMPFS=no
 
 select_mirror () {
     case $ARCH in
@@ -118,11 +171,14 @@ EXCLUDE=""
 LIST=""
 SUBARCH=""
 PROPOSED=""
+IMAGEFORMAT="squashfs"
 # must be in the "team / PPA name" form; e.g. "moblin/ppa"; the default PPA
 # name is "ppa", don't omit it
 PPA=""
+APT_SOURCE=""
+FORCE_YES=""
 
-while getopts :d:e:i:I:m:S:s:a:p name; do case $name in
+while getopts :d:e:i:I:m:S:s:a:A:f:p name; do case $name in
     d)  STE=$OPTARG;;
     e)  EXCLUDE="$EXCLUDE $OPTARG";;
     i)  LIST="$LIST $OPTARG";;
@@ -131,6 +187,8 @@ while getopts :d:e:i:I:m:S:s:a:p name; do case $name in
     S)	USZ="$OPTARG";;
     s)	SUBARCH="$OPTARG";;
     a)	ARCH="$OPTARG";;
+    A)	APT_SOURCE="$OPTARG";;
+    f)	IMAGEFORMAT="$OPTARG";;
     p)  PROPOSED="yes";;
     \?) echo bad usage >&2; exit 2;;
     \:) echo missing argument >&2; exit 2;;
@@ -138,15 +196,21 @@ esac; done;
 shift $((OPTIND-1))
 
 if (( $# == 0 )) || [ "X$1" = "Xall" ]; then
-    set -- ubuntu kubuntu kubuntu-netbook edubuntu xubuntu mythbuntu gobuntu base
+    set -- ubuntu kubuntu kubuntu-mobile kubuntu-netbook edubuntu xubuntu mythbuntu gobuntu base
     if [ "$ARCH" = "i386" ]; then
-        set -- ubuntu ubuntu-dvd kubuntu kubuntu-dvd kubuntu-netbook edubuntu edubuntu-dvd mythbuntu xubuntu gobuntu base
+        set -- ubuntu ubuntu-dvd kubuntu kubuntu-dvd kubuntu-mobile kubuntu-netbook edubuntu edubuntu-dvd mythbuntu xubuntu gobuntu base
     fi
 fi
 
+if [ "$IMAGEFORMAT" = "ext2" ] || [ "$IMAGEFORMAT" = "ext3" ]; then
+    LIVE_BOOT_SCRIPTS="jasper"
+else
+    LIVE_BOOT_SCRIPTS="casper lupin-casper"
+fi
+
 for arg in "$@"; do
     case "$arg" in
-       ubuntu|ubuntu-dvd|ubuntu-lpia|edubuntu|edubuntu-dvd|kubuntu|kubuntu-dvd|kubuntu-netbook|xubuntu|mythbuntu|gobuntu|ubuntu-mid|ubuntu-netbook|ubuntu-moblin-remix|base|tocd)
+       ubuntu|ubuntu-dvd|ubuntu-lpia|edubuntu|edubuntu-dvd|kubuntu|kubuntu-dvd|kubuntu-mobile|kubuntu-netbook|xubuntu|mythbuntu|gobuntu|ubuntu-mid|ubuntu-netbook|ubuntu-moblin-remix|base|tocd)
 	    ;;
 	*)
 	    echo bad name >&2;
@@ -167,7 +231,13 @@ for FS in "$@"; do
     # Just in case there's some leftover junk here:
     cleanup 2>/dev/null || true
 
+    umount ${ROOT} || true
     rm -rf ${ROOT}
+    mkdir ${ROOT}
+    # maybe use a tmpfs
+    if  test yes = "$TMPFS"; then
+        mount -t tmpfs -o size=8192M tmpfs ${ROOT} && echo using tmpfs
+    fi
 
     mkdir -p ${ROOT}var/cache/debconf
     cat << @@EOF > ${ROOT}var/cache/debconf/config.dat
@@ -182,29 +252,34 @@ Flags: seen
     case "$FS" in
 	ubuntu|ubuntu-lpia|ubuntu-dvd)
 	    LIST="$LIST minimal^ standard^ ubuntu-desktop^"
-	    LIVELIST="ubuntu-live^ laptop-detect casper lupin-casper"
+	    LIVELIST="ubuntu-live^ laptop-detect $LIVE_BOOT_SCRIPTS"
 	    ;;
 	kubuntu|kubuntu-dvd)
 	    LIST="$LIST minimal^ standard^ kubuntu-desktop^"
-	    LIVELIST="kubuntu-live^ laptop-detect casper lupin-casper"
+	    LIVELIST="kubuntu-live^ laptop-detect $LIVE_BOOT_SCRIPTS"
+	    ;;
+	kubuntu-mobile)
+	    LIST="$LIST minimal^ standard^ kubuntu-mobile^"
+	    LIVELIST="language-support-en kubuntu-mobile-live^ laptop-detect $LIVE_BOOT_SCRIPTS"
+	    COMP="main restricted universe"
 	    ;;
 	kubuntu-netbook)
 	    LIST="$LIST minimal^ standard^ kubuntu-netbook^"
-	    LIVELIST="language-support-en kubuntu-netbook-live^ laptop-detect casper lupin-casper"
+	    LIVELIST="language-support-en kubuntu-netbook-live^ laptop-detect $LIVE_BOOT_SCRIPTS"
 	    ;;
 	edubuntu|edubuntu-dvd)
 	    LIST="$LIST minimal^ standard^ edubuntu-desktop-gnome^"
-	    LIVELIST="edubuntu-live^ laptop-detect casper lupin-casper"
+	    LIVELIST="edubuntu-live^ laptop-detect $LIVE_BOOT_SCRIPTS"
 	    COMP="main restricted universe"
 	    ;;
 	xubuntu)
 	    LIST="$LIST minimal^ standard^ xterm xubuntu-desktop^"
-	    LIVELIST="xubuntu-live^ laptop-detect casper lupin-casper"
+	    LIVELIST="xubuntu-live^ laptop-detect $LIVE_BOOT_SCRIPTS"
 	    COMP="main restricted universe multiverse"
 	    ;;
 	gobuntu)
 	    LIST="$LIST minimal^ standard^ gobuntu-desktop^"
-	    LIVELIST="gobuntu-live^ laptop-detect casper lupin-casper"
+	    LIVELIST="gobuntu-live^ laptop-detect $LIVE_BOOT_SCRIPTS"
 	    COMP="main"
 	    ;;
 	ubuntu-mid)
@@ -214,11 +289,11 @@ Flags: seen
 	    ;;
 	ubuntu-netbook)
 	    LIST="$LIST minimal^ standard^ ubuntu-netbook^"
-	    LIVELIST="netbook-live^ laptop-detect casper lupin-casper"
+	    LIVELIST="netbook-live^ laptop-detect $LIVE_BOOT_SCRIPTS"
 	    ;;
 	mythbuntu)
 	    LIST="$LIST minimal^ standard^ mythbuntu-desktop^"
-	    LIVELIST="mythbuntu-live^ laptop-detect casper lupin-casper"
+	    LIVELIST="mythbuntu-live^ laptop-detect $LIVE_BOOT_SCRIPTS"
 	    COMP="main restricted universe multiverse"
 	    ;;
 	ubuntu-moblin-remix)
@@ -229,7 +304,7 @@ Flags: seen
 	    ;;
 	base)
 	    LIST="$LIST minimal^ standard^"
-	    LIVELIST="casper lupin-casper"
+	    LIVELIST="$LIVE_BOOT_SCRIPTS"
 	    ;;
 	tocd)
 	    LIST="$LIST minimal^ standard^"
@@ -268,6 +343,7 @@ Flags: seen
     esac
 
     dpkg -l livecd-rootfs || true	# get our version # in the log.
+    checkpoint "Bootstrapping base system"
     debootstrap --components=$(echo $COMP | sed 's/ /,/g') --arch $TARGETARCH $STE $ROOT $MIRROR
 
     # Recent dpkg has started complaining pretty loudly if dev/pts isn't 
@@ -352,6 +428,10 @@ link_in_boot = $link_in_boot
 					LIST="$LIST linux-omap"
 					STRIP_VMLINUZ=no
 					;;
+				omap4)
+					LIST="$LIST linux-omap4"
+					STRIP_VMLINUZ=no
+					;;
 			esac;;
 	*)		echo "Unknown architecture: no kernel."; exit 1;;
     esac
@@ -368,11 +448,12 @@ link_in_boot = $link_in_boot
 	cat << @@EOF > ${ROOT}etc/apt/preferences
 Package: ssl-cert
 Pin: version 1.0.14-0ubuntu2
-Pin-Priority: 900
+Pin-Priority: 991
 @@EOF
     fi
 
     # Create a good sources.list, and finish the install
+    checkpoint "Configuring APT"
     echo deb $MIRROR $STE ${COMP} > ${ROOT}etc/apt/sources.list
     echo deb $MIRROR ${STE}-security ${COMP} >> ${ROOT}etc/apt/sources.list
     echo deb $MIRROR ${STE}-updates ${COMP} >> ${ROOT}etc/apt/sources.list
@@ -393,6 +474,24 @@ Pin: release o=LP-PPA-$origin
 Pin-Priority: 550
 @@EOF
     fi
+    if [ -n "$APT_SOURCE" ]; then
+	echo deb $APT_SOURCE $STE $COMP >> ${ROOT}etc/apt/sources.list
+	# allow unsigned sources, unfortunately
+	mkdir -p ${ROOT}etc/apt/apt.conf.d
+	echo 'APT::Get::AllowUnauthenticated "yes";' > ${ROOT}etc/apt/apt.conf.d/extra-source-allows-unauthenticated-sources
+	# and add the preferences rules :(
+	APT_PIN=$(echo $APT_SOURCE | sed -e's,http://,,; s,/.*,,')
+	cat >> ${ROOT}etc/apt/preferences << @@EOF
+Package: *
+Pin: release a=$STE
+Pin-Priority: 500
+
+Package: *
+Pin: origin $APT_PIN
+Pin-Priority: 900
+@@EOF
+	FORCE_YES=--force-yes
+    fi
 
     if [ "$FS" = "ubuntu-moblin-remix" ]; then
 	chroot $ROOT apt-get update
@@ -408,13 +507,21 @@ Pin-Priority: 550
     cp ${ROOT}etc/apt/trusted.gpg ${ROOT}etc/apt/trusted.gpg.$$
     cat /etc/apt/trusted.gpg >> ${ROOT}etc/apt/trusted.gpg
 
+    # update and immediately restore the trusted keyring
     chroot $ROOT apt-get update
-    chroot $ROOT apt-get -y --purge dist-upgrade </dev/null
+    # we restore the keyring here because on dist-upgrade the
+    # ubuntu-extras-keyring package installs additional keys
+    mv ${ROOT}etc/apt/trusted.gpg.$$ ${ROOT}etc/apt/trusted.gpg
+
+    checkpoint "Upgrading"
+    chroot $ROOT apt-get -y $FORCE_YES --purge dist-upgrade </dev/null
+    checkpoint "Installing main packages"
     chroot $ROOT apt-get -y --purge install $LIST </dev/null
 
     # launchpad likes to put dependencies of seeded packages in tasks along with the
     # actual seeded packages.  In general, this isn't an issue.  With updated kernels
     # and point-releases, though, we end up with extra header packages:
+    checkpoint "Cleaning up kernel headers"
     chroot ${ROOT} dpkg -l linux-headers-2\* | grep ^i | awk '{print $2}' \
         > livecd.${FSS}.manifest-headers
     chroot ${ROOT} dpkg -l linux-headers-\* | grep ^i | awk '{print $2}' \
@@ -424,6 +531,11 @@ Pin-Priority: 550
     for i in `comm -3 livecd.${FSS}.manifest-headers livecd.${FSS}.manifest-headers-full`; do
         HEADERMETA="$HEADERMETA $i"
     done
+    case $TARGETARCH in
+        armel)
+            HEADERMETA=linux-headers-${SUBARCH}
+            ;;
+    esac
     rm -f livecd.${FSS}.manifest-headers livecd.${FSS}.manifest-headers-full
     chroot ${ROOT} apt-get -y --purge remove $HEADERPACKAGES </dev/null || true
     chroot ${ROOT} apt-get -y --purge install $HEADERMETA </dev/null || true
@@ -431,6 +543,7 @@ Pin-Priority: 550
 
     chroot ${ROOT} dpkg-query -W --showformat='${Package} ${Version}\n' \
 	> livecd.${FSS}.manifest-desktop
+    checkpoint "Installing live packages"
     chroot $ROOT apt-get -y --purge install $LIVELIST </dev/null
     case $FS in
 	edubuntu)
@@ -445,6 +558,9 @@ Pin-Priority: 550
     esac
     chroot ${ROOT} dpkg-query -W --showformat='${Package} ${Version}\n' \
 	> livecd.${FSS}.manifest
+
+    checkpoint "Cleaning up"
+
     kill_users
 
     chroot $ROOT /etc/cron.daily/mlocate || true
@@ -534,7 +650,6 @@ Pin: release o=LP-PPA-$origin
 Pin-Priority: 550
 @@EOF
     fi
-    mv ${ROOT}etc/apt/trusted.gpg.$$ ${ROOT}etc/apt/trusted.gpg
 
     # get rid of the .debs - we don't need them.
     chroot ${ROOT} apt-get clean
@@ -585,10 +700,18 @@ Pin-Priority: 550
     esac
 
     # No point keeping Gnome icon cache around for Kubuntu
-    if [ "$FS" = "kubuntu" ] || [ "$FS" = "kubuntu-netbook" ]; then
+    if [ "$FS" = "kubuntu" ] || \
+       [ "$FS" = "kubuntu-netbook" ] || \
+       [ "$FS" = "kubuntu-mobile" ]; then
         rm -f ${ROOT}/usr/share/icons/*/icon-theme.cache
     fi
 
+    if [ "$FS" = ubuntu-netbook ] && \
+       [ "$TARGETARCH+$SUBARCH" = armel+dove ] && \
+       [ -e "$ROOT/usr/share/xsessions/une-efl.desktop" ]; then
+        chroot $ROOT /usr/lib/gdm/gdm-set-default-session une-efl
+    fi
+
     # all done with the chroot; reset the debconf frontend, so Colin doesn't cry
     echo RESET debconf/frontend | chroot $ROOT debconf-communicate
     echo FSET debconf/frontend seen true | chroot $ROOT debconf-communicate
@@ -597,12 +720,14 @@ Pin-Priority: 550
     rm -f ${ROOT}/var/cache/debconf/*-old
 
     # show the size of directories in /usr/share/doc
+    checkpoint "Checking size of /usr/share/doc"
     echo BEGIN docdirs
     (cd $ROOT && find usr/share/doc -maxdepth 1 -type d | xargs du -s | sort -nr)
     echo END docdirs
 
     # search for duplicate files, write the summary to stdout, 
     if which fdupes >/dev/null 2>&1; then
+	checkpoint "Checking for duplicate files"
 	echo "first line: <total size for dupes> <different dupes> <all dupes>"
 	echo "data lines: <size for dupes> <number of dupes> <file size> <filename> [<filename> ...]"
 	echo BEGIN fdupes
@@ -618,40 +743,39 @@ Pin-Priority: 550
     perl -i -nle 'print unless /^Package: language-(pack|support)/ .. /^$/;' \
         ${ROOT}/var/lib/apt/extended_states
 
-  # And run the cleanup function dead last, to umount /proc after nothing
-  # else needs to be run in the chroot (umounting it earlier breaks rm):
-  cleanup
+    # And run the cleanup function dead last, to umount /proc after nothing
+    # else needs to be run in the chroot (umounting it earlier breaks rm):
+    cleanup
 
-  # Squashfs does not report unpacked disk space usage, which is explained at
-  # <http://lkml.org/lkml/2006/6/16/163>.  However, we would like to cache this
-  # number for partman's sufficient free space check and ubiquity's
-  # installation progress calculation.
-  printf $(du -sx --block-size=1 ${ROOT} | cut -f1) > livecd.${FSS}.size || true
+    # Squashfs does not report unpacked disk space usage, which is explained at
+    # <http://lkml.org/lkml/2006/6/16/163>.  However, we would like to cache this
+    # number for partman's sufficient free space check and ubiquity's
+    # installation progress calculation.
+    checkpoint "Calculating total size"
+    printf $(du -sx --block-size=1 ${ROOT} | cut -f1) > livecd.${FSS}.size || true
 
-  livefs_squash()
-  {
-    squashsort="http://people.ubuntu.com/~tfheen/livesort/${FSS}.list.${TARGETARCH}"
-    #if wget -O livecd.${FSS}.sort ${squashsort} > /dev/null 2>&1; then
-    if false; then
-      echo "Using the squashfs sort list from ${squashsort}."
+
+    # Build our images
+    checkpoint "Building image"
+    if [ "$IMAGEFORMAT" = "ext2" ] || [ "$IMAGEFORMAT" = "ext3" ]; then
+        livefs_ext2
     else
-      echo "Unable to fetch squashfs sort list; using a blank list."
-      : > livecd.${FSS}.sort
+        livefs_squash
     fi
 
-    # make sure there is no old squashfs idling around
-    rm -f livecd.${FSS}.squashfs
-
-    mksquashfs ${ROOT} livecd.${FSS}.squashfs -sort livecd.${FSS}.sort
-    chmod 644 livecd.${FSS}.squashfs
-  }
-
-  livefs_squash
+    # Upgrade ext2->ext3 if that's what is requested
+    if [ "$IMAGEFORMAT" = "ext3" ]; then
+        tune2fs -j livecd.${FSS}.ext2
+        mv livecd.${FSS}.ext2 livecd.${FSS}.ext3
+        # temporary workaround for LP: #583317 with ext3 images
+        e2fsck -fy livecd.${FSS}.ext3 || true
+    fi
 
     # LTSP chroot building (only in 32bit and for Edubuntu (DVD))
     case $FS in
         edubuntu-dvd)
             if [ "$TARGETARCH" = "i386" ]; then
+                checkpoint "Building LTSP chroot"
                 ltsp-build-client --base $(pwd) --mirror $MIRROR --arch $TARGETARCH --dist $STE --chroot ltsp-live --purge-chroot --skipimage
                 mkdir -p $(pwd)/images
                 mksquashfs $(pwd)/ltsp-live $(pwd)/images/ltsp-live.img -noF -noD -noI -no-exports -e cdrom
@@ -667,3 +791,5 @@ Pin-Priority: 550
         ;;
     esac
 done
+
+checkpoint "Done"
